"""Experimental data for the dam break problem.

The data is extracted from:

Martin and Moyce 1952
"An Experimental Study of the Collapse of Liquid Columns on a Rigid Horizontal
Plane", J. C. Martin and W. J. Moyce, Philososphical Transactions of the Royal
Society of London Series A, 244, 312--324 (1952).

and

"Moving-Particle Semi-Implicit Method for Fragmentation of Incompressible
fluid", S. Koshizuka and Y. Oka, Nuclear Science and Engineering, 123,
421--434 (1996).

"""

import numpy as np
from io import StringIO

# This is the data for n^2=2 and a=1.125 from Figure 3.
mm_data_1 = u"""
0.849   1.245
1.212   1.443
1.602   1.884
2.283   2.689
2.950   3.728
3.598   4.528
3.905   4.999
4.592   5.841
4.961   6.271
5.316   6.717
"""

# This is the data for n^2=2 and a=2.25 from Figure 3.
mm_data_2 = u"""
0.832   1.217
1.219   1.474
1.997   2.292
2.547   2.995
3.345   4.134
4.034   4.944
4.418   5.881
5.091   6.980
5.685   7.945
6.306   8.966
6.822   9.986
7.439   10.963
8.031   11.977
8.633   13.005
9.237   13.970
"""

ko_data = u"""
0.0     1.000
0.381   1.111
0.769   1.252
1.153   1.505
1.537   1.892
1.935   2.241
2.323   2.615
2.719   3.003
3.096   3.624
"""

def get_martin_moyce_1():
    """Returns t*sqrt(2*g/a), z/a for the case where a = 1.125 inches
    """
    # z/a vs t*np.sqrt(2*g/a)
    t, z = np.loadtxt(StringIO(mm_data_1), unpack=True)
    return t, z

def get_martin_moyce_2():
    """Returns t*sqrt(2*g/a), z/a for the case where a = 2.25 inches
    """
    # z/a vs t*np.sqrt(2*g/a)
    t, z = np.loadtxt(StringIO(mm_data_2), unpack=True)
    return t, z


def get_koshizuka_oka_data():
    # z/L vs t*np.sqrt(2*g/L)
    t, z = np.loadtxt(StringIO(ko_data), unpack=True)
    return t, z



"""Convenience script for running various PySPH related tasks.
"""

from __future__ import print_function

from argparse import ArgumentParser
from os.path import exists, join
import sys

def run_viewer(args):
    from pysph.tools.mayavi_viewer import main
    main(args)

def run_examples(args):
    from pysph.examples.run import main
    main(args)

def output_vtk(args):
    from pysph.solver.vtk_output import main
    main(args)

def _has_pysph_dir():
    init_py = join('pysph', '__init__.py')
    init_pyc = join('pysph', '__init__.pyc')
    return exists(init_py) or exists(init_pyc)

def run_tests(args):
    wd = ['-w', 'docs'] if _has_pysph_dir() else []
    argv = ['nosetests', 'pysph', 'pyzoltan'] + wd + args
    from nose.core import main
    main(argv=argv)

def main():
    parser = ArgumentParser(description=__doc__, add_help=False)
    parser.add_argument(
        "-h", "--help", action="store_true", default=False, dest="help",
        help="show this help message and exit"
    )
    subparsers = parser.add_subparsers(help='sub-command help')

    viewer = subparsers.add_parser(
        'view', help='View output files generated by PySPH',
        add_help=False
    )
    viewer.set_defaults(func=run_viewer)

    runner = subparsers.add_parser(
        'run', help='Run PySPH examples',
        add_help=False
    )
    runner.set_defaults(func=run_examples)

    vtk_out = subparsers.add_parser(
        'dump_vtk', help='Dump VTK Output',
         add_help=False
    )
    vtk_out.set_defaults(func=output_vtk)
    tests = subparsers.add_parser(
        'test', help='Run entire PySPH test-suite',
        add_help=False
    )
    tests.set_defaults(func=run_tests)

    if len(sys.argv) == 1 or \
        (len(sys.argv) > 1 and sys.argv[1] in ['-h', '--help']):
        parser.print_help()
        sys.exit()

    args, extra = parser.parse_known_args()
    args.func(extra)

if __name__ == '__main__':
    main()



"""The basic equations for the IISPH formulation of

    M. Ihmsen, J. Cornelis, B. Solenthaler, C. Horvath, M. Teschner, "Implicit
    Incompressible SPH," IEEE Transactions on Visualization and Computer
    Graphics, vol. 20, no. 3, pp. 426-435, March 2014.
    http://dx.doi.org/10.1109/TVCG.2013.105

"""

from numpy import sqrt
from pysph.sph.equation import Equation
from pysph.sph.integrator_step import IntegratorStep
from pysph.base.reduce_array import serial_reduce_array, parallel_reduce_array


class IISPHStep(IntegratorStep):
    """A straightforward and simple integrator to be used for IISPH.
    """
    def initialize(self):
        pass

    def stage1(self, d_idx, d_x, d_y, d_z, d_u, d_v, d_w,
               d_uadv, d_vadv, d_wadv, d_au, d_av, d_aw,
               d_ax, d_ay, d_az, dt):
        d_u[d_idx] = d_uadv[d_idx] + dt * d_au[d_idx]
        d_v[d_idx] = d_vadv[d_idx] + dt * d_av[d_idx]
        d_w[d_idx] = d_wadv[d_idx] + dt * d_aw[d_idx]

        d_x[d_idx] += dt * d_u[d_idx]
        d_y[d_idx] += dt * d_v[d_idx]
        d_z[d_idx] += dt * d_w[d_idx]


class NumberDensity(Equation):
    def initialize(self, d_idx, d_V):
        d_V[d_idx] = 0.0

    def loop(self, d_idx, d_V, WIJ):
        d_V[d_idx] += WIJ

class SummationDensity(Equation):
    def initialize(self, d_idx, d_rho):
        d_rho[d_idx] = 0.0

    def loop(self, d_idx, d_rho, s_idx, s_m, WIJ):
        d_rho[d_idx] += s_m[s_idx]*WIJ

class SummationDensityBoundary(Equation):
    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(SummationDensityBoundary, self).__init__(dest, sources)

    def loop(self, d_idx, d_rho, s_idx, s_V, WIJ):
        d_rho[d_idx] += self.rho0/s_V[s_idx]*WIJ

class NormalizedSummationDensity(Equation):
    def initialize(self, d_idx, d_rho, d_rho_adv, d_rho0, d_V):
        d_rho0[d_idx] = d_rho[d_idx]
        d_rho[d_idx] = 0.0
        d_rho_adv[d_idx] = 0.0
        d_V[d_idx] = 0.0

    def loop(self, d_idx, d_rho, d_rho_adv, d_V, s_idx, s_m, s_rho0, WIJ):
        tmp = s_m[s_idx]*WIJ
        d_rho[d_idx] += tmp
        d_rho_adv[d_idx] += tmp/s_rho0[s_idx]
        d_V[d_idx] += WIJ

    def post_loop(self, d_idx, d_rho, d_rho_adv):
        d_rho[d_idx] /= d_rho_adv[d_idx]


class AdvectionAcceleration(Equation):
    def __init__(self, dest, sources, gx=0.0, gy=0.0, gz=0.0):
        self.gx = gx
        self.gy = gy
        self.gz = gz
        super(AdvectionAcceleration, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw, d_uadv, d_vadv, d_wadv):
        d_au[d_idx] = self.gx
        d_av[d_idx] = self.gy
        d_aw[d_idx] = self.gz
        d_uadv[d_idx] = 0.0
        d_vadv[d_idx] = 0.0
        d_wadv[d_idx] = 0.0

    def loop(self):
        pass

    def post_loop(self, d_idx, d_au, d_av, d_aw, d_uadv, d_vadv, d_wadv,
                  d_u, d_v, d_w, dt=0.0):
        d_uadv[d_idx] = d_u[d_idx] + dt*d_au[d_idx]
        d_vadv[d_idx] = d_v[d_idx] + dt*d_av[d_idx]
        d_wadv[d_idx] = d_w[d_idx] + dt*d_aw[d_idx]

class ViscosityAcceleration(Equation):
    def __init__(self, dest, sources, nu):
        self.nu = nu
        super(ViscosityAcceleration, self).__init__(dest, sources)

    def loop(self, d_idx, d_au, d_av, d_aw, s_idx, s_m, EPS,
             VIJ, XIJ, RHOIJ1, R2IJ, DWIJ):
        dwijdotxij = DWIJ[0]*XIJ[0] + DWIJ[1]*XIJ[1] + DWIJ[2]*XIJ[2]
        fac = 2.0*self.nu*s_m[s_idx]*RHOIJ1*dwijdotxij/(R2IJ + EPS)
        d_au[d_idx] += fac*VIJ[0]
        d_av[d_idx] += fac*VIJ[1]
        d_aw[d_idx] += fac*VIJ[2]

class ViscosityAccelerationBoundary(Equation):
    """The acceleration on the fluid due to a boundary.
    """
    def __init__(self, dest, sources, rho0, nu):
        self.nu = nu
        self.rho0 = rho0
        super(ViscosityAccelerationBoundary, self).__init__(dest, sources)

    def loop(self, d_idx, d_au, d_av, d_aw, d_rho, s_idx, s_V, EPS,
             VIJ, XIJ, R2IJ, DWIJ):
        phi_b = self.rho0/(s_V[s_idx]*d_rho[d_idx])
        dwijdotxij = DWIJ[0]*XIJ[0] + DWIJ[1]*XIJ[1] + DWIJ[2]*XIJ[2]

        fac = 2.0*self.nu*phi_b*dwijdotxij/(R2IJ + EPS)
        d_au[d_idx] += fac*VIJ[0]
        d_av[d_idx] += fac*VIJ[1]
        d_aw[d_idx] += fac*VIJ[2]


class ComputeDII(Equation):
    def initialize(self, d_idx, d_dii0, d_dii1, d_dii2):
        d_dii0[d_idx] = 0.0
        d_dii1[d_idx] = 0.0
        d_dii2[d_idx] = 0.0

    def loop(self, d_idx, d_rho, d_dii0, d_dii1, d_dii2,
             s_idx, s_m, DWIJ, dt=0.0):
        rho_1 = 1.0/d_rho[d_idx]
        fac = -dt*dt*s_m[s_idx]*rho_1*rho_1
        d_dii0[d_idx] += fac*DWIJ[0]
        d_dii1[d_idx] += fac*DWIJ[1]
        d_dii2[d_idx] += fac*DWIJ[2]


class ComputeDIIBoundary(Equation):
    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(ComputeDIIBoundary, self).__init__(dest, sources)

    def loop(self, d_idx, d_dii0, d_dii1, d_dii2, d_rho,
             s_idx, s_m, s_V, DWIJ, dt=0.0):
        rhoi1 = 1.0/d_rho[d_idx]
        fac = -dt*dt*rhoi1*rhoi1*self.rho0/s_V[s_idx]
        d_dii0[d_idx] += fac*DWIJ[0]
        d_dii1[d_idx] += fac*DWIJ[1]
        d_dii2[d_idx] += fac*DWIJ[2]


class ComputeRhoAdvection(Equation):
    def initialize(self, d_idx, d_rho_adv, d_rho, d_p0, d_p, d_piter, d_aii):
        d_rho_adv[d_idx] = d_rho[d_idx]
        d_p0[d_idx] = d_p[d_idx]
        d_piter[d_idx] = 0.5*d_p[d_idx]

    def loop(self, d_idx, d_rho, d_rho_adv, d_uadv, d_vadv, d_wadv, d_u, d_v, d_w,
             s_idx, s_m, s_uadv, s_vadv, s_wadv, DWIJ, dt=0.0):

        vijdotdwij = (d_uadv[d_idx] - s_uadv[s_idx])*DWIJ[0] + \
                     (d_vadv[d_idx] - s_vadv[s_idx])*DWIJ[1] + \
                     (d_wadv[d_idx] - s_wadv[s_idx])*DWIJ[2]

        d_rho_adv[d_idx] += dt*s_m[s_idx]*vijdotdwij


class ComputeRhoBoundary(Equation):
    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(ComputeRhoBoundary, self).__init__(dest, sources)

    def loop(self, d_idx, d_rho, d_rho_adv, d_uadv, d_vadv, d_wadv,
             s_idx, s_u, s_v, s_w, s_V, WIJ, DWIJ, dt=0.0):
        phi_b = self.rho0/s_V[s_idx]

        vijdotdwij = (d_uadv[d_idx] - s_u[s_idx])*DWIJ[0] + \
                     (d_vadv[d_idx] - s_v[s_idx])*DWIJ[1] + \
                     (d_wadv[d_idx] - s_w[s_idx])*DWIJ[2]
        d_rho_adv[d_idx] += dt*phi_b*vijdotdwij


class ComputeAII(Equation):
    def initialize(self, d_idx, d_aii):
        d_aii[d_idx] = 0.0

    def loop(self, d_idx, d_aii, d_dii0, d_dii1, d_dii2, d_m, d_rho,
             s_idx, s_m, s_rho, DWIJ, dt=0.0):
        rho1 = 1.0/d_rho[d_idx]
        fac = dt*dt*d_m[d_idx]*rho1*rho1
        # The following is m_j (d_ii - d_ji) . DWIJ
        # DWIJ = -DWJI
        dijdotdwij = (d_dii0[d_idx] - fac*DWIJ[0])*DWIJ[0] + \
                     (d_dii1[d_idx] - fac*DWIJ[1])*DWIJ[1] + \
                     (d_dii2[d_idx] - fac*DWIJ[2])*DWIJ[2]
        d_aii[d_idx] += s_m[s_idx]*dijdotdwij


class ComputeAIIBoundary(Equation):
    """ This is important and not really discussed in the original IISPH paper.
    """
    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(ComputeAIIBoundary, self).__init__(dest, sources)

    def loop(self, d_idx, d_aii, d_dii0, d_dii1, d_dii2, d_rho,
             s_idx, s_m, s_V, DWIJ, dt=0.0):
        phi_b = self.rho0/s_V[s_idx]
        dijdotdwij = d_dii0[d_idx]*DWIJ[0] + d_dii1[d_idx]*DWIJ[1] + \
                     d_dii2[d_idx]*DWIJ[2]
        d_aii[d_idx] += phi_b*dijdotdwij


class ComputeDIJPJ(Equation):
    def initialize(self, d_idx, d_dijpj0, d_dijpj1, d_dijpj2):
        d_dijpj0[d_idx] = 0.0
        d_dijpj1[d_idx] = 0.0
        d_dijpj2[d_idx] = 0.0

    def loop(self, d_idx, d_dijpj0, d_dijpj1, d_dijpj2,
             s_idx, s_m, s_rho, s_piter, DWIJ, dt=0.0):
        rho1 = 1.0/s_rho[s_idx]
        fac = -dt*dt*s_m[s_idx]*rho1*rho1*s_piter[s_idx]
        d_dijpj0[d_idx] += fac*DWIJ[0]
        d_dijpj1[d_idx] += fac*DWIJ[1]
        d_dijpj2[d_idx] += fac*DWIJ[2]


class PressureSolve(Equation):
    def __init__(self, dest, sources, rho0, omega=0.5,
                 tolerance=1e-2, debug=False):
        self.rho0 = rho0
        self.omega = omega
        self.compression = 0.0
        self.debug = debug
        self.tolerance = tolerance
        super(PressureSolve, self).__init__(dest, sources)

    def initialize(self, d_idx, d_p, d_compression):
        d_p[d_idx] = 0.0
        d_compression[d_idx] = 0.0

    def loop(self, d_idx, d_p, d_piter, d_rho, d_m, d_dijpj0, d_dijpj1, d_dijpj2,
             s_idx, s_m, s_dii0, s_dii1, s_dii2,
             s_piter, s_dijpj0, s_dijpj1, s_dijpj2, DWIJ, dt=0.0):

        # Note that a good way to check this is to see that when
        # d_idx == s_idx the contribution is zero, as is expected.
        rho1 = 1.0/d_rho[d_idx]
        fac = dt*dt*d_m[d_idx]*rho1*rho1*d_piter[d_idx]
        djkpk0 = s_dijpj0[s_idx] - fac*DWIJ[0]
        djkpk1 = s_dijpj1[s_idx] - fac*DWIJ[1]
        djkpk2 = s_dijpj2[s_idx] - fac*DWIJ[2]

        tmp0 = d_dijpj0[d_idx] - s_dii0[s_idx]*s_piter[s_idx] - djkpk0
        tmp1 = d_dijpj1[d_idx] - s_dii1[s_idx]*s_piter[s_idx] - djkpk1
        tmp2 = d_dijpj2[d_idx] - s_dii2[s_idx]*s_piter[s_idx] - djkpk2
        tmpdotdwij = tmp0*DWIJ[0] + tmp1*DWIJ[1] + tmp2*DWIJ[2]

        # This is corrected in the post_loop.
        d_p[d_idx] += s_m[s_idx]*tmpdotdwij

    def post_loop(self, d_idx, d_piter, d_p0, d_p, d_aii, d_rho_adv, d_rho,
                  d_compression, dt=0.0):
        #tmp = d_rho[d_idx] - d_rho_adv[d_idx] - d_p[d_idx]
        tmp = self.rho0 - d_rho_adv[d_idx] - d_p[d_idx]
        p = (1.0 - self.omega)*d_piter[d_idx] + self.omega/d_aii[d_idx]*tmp

        aii_min = dt*dt*0.01

        # Clamp pressure to positive values.
        if p < 0.0:
            p = 0.0
        elif abs(d_aii[d_idx]) < aii_min :
            p = 0.0
        else:
            d_compression[d_idx] = abs(p*d_aii[d_idx] - tmp)

        d_piter[d_idx] = p
        d_p[d_idx] = p

    def reduce(self, dst):
        dst.tmp_comp[0] = serial_reduce_array(dst.array.compression > 0.0, 'sum')
        dst.tmp_comp[1] = serial_reduce_array(dst.array.compression, 'sum')
        dst.tmp_comp.set_data(parallel_reduce_array(dst.tmp_comp, 'sum'))
        if dst.tmp_comp[0] > 0:
            comp = dst.tmp_comp[1]/dst.tmp_comp[0]/self.rho0
        else:
            comp = 0.0
        self.compression = comp

    def converged(self):
        debug = self.debug
        compression = self.compression

        if compression > self.tolerance:
            if debug:
                print("Not converged:", compression)
            return -1.0
        else:
            if debug:
                print("Converged:", compression)
            return 1.0

class PressureSolveBoundary(Equation):
    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(PressureSolveBoundary, self).__init__(dest, sources)

    def loop(self, d_idx, d_p, d_rho, d_dijpj0, d_dijpj1, d_dijpj2,
             s_idx, s_V, DWIJ):
        phi_b = self.rho0/s_V[s_idx]
        dijdotwij = d_dijpj0[d_idx]*DWIJ[0] + \
                    d_dijpj1[d_idx]*DWIJ[1] + \
                    d_dijpj2[d_idx]*DWIJ[2]
        d_p[d_idx] += phi_b*dijdotwij


class PressureForce(Equation):
    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, d_rho, d_p, d_au, d_av, d_aw,
             s_idx, s_m, s_rho, s_p, DWIJ):
        rhoi1 = 1.0/d_rho[d_idx]
        rhoj1 = 1.0/s_rho[s_idx]
        fac = -s_m[s_idx]*(d_p[d_idx]*rhoi1*rhoi1 + s_p[s_idx]*rhoj1*rhoj1)
        d_au[d_idx] += fac*DWIJ[0]
        d_av[d_idx] += fac*DWIJ[1]
        d_aw[d_idx] += fac*DWIJ[2]

    def post_loop(self, d_idx, d_au, d_av, d_aw,
                  d_uadv, d_vadv, d_wadv, DT_ADAPT):
        fac = d_au[d_idx]*d_au[d_idx] + d_av[d_idx]*d_av[d_idx] +\
                   d_aw[d_idx]*d_aw[d_idx]
        vmag = sqrt(d_uadv[d_idx]*d_uadv[d_idx] + d_vadv[d_idx]*d_vadv[d_idx] +
                    d_wadv[d_idx]*d_wadv[d_idx])
        DT_ADAPT[0] = max(2.0*vmag, DT_ADAPT[0])
        DT_ADAPT[1] = max(2.0*fac, DT_ADAPT[1])


class PressureForceBoundary(Equation):
    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(PressureForceBoundary, self).__init__(dest, sources)

    def loop(self, d_idx, d_rho, d_au, d_av, d_aw,  d_p, s_idx, s_V, DWIJ):
        rho1 = 1.0/d_rho[d_idx]
        fac = -d_p[d_idx]*rho1*rho1*self.rho0/s_V[s_idx]
        d_au[d_idx] += fac*DWIJ[0]
        d_av[d_idx] += fac*DWIJ[1]
        d_aw[d_idx] += fac*DWIJ[2]



"""Two-dimensional dam break over a dry bed.  (30 minutes)

The case is described in "State of the art classical SPH for free surface
flows", Moncho Gomez-Gesteira, Benedict D Rogers, Robert A, Dalrymple and Alex
J.C Crespo, Journal of Hydraulic Research, Vol 48, Extra Issue (2010), pp
6-27. DOI:10.1080/00221686.2010.9641242

"""

import os
import numpy as np

from pysph.base.kernels import WendlandQuintic
from pysph.examples._db_geometry import DamBreak2DGeometry
from pysph.solver.application import Application
from pysph.sph.scheme import WCSPHScheme

fluid_column_height = 2.0
fluid_column_width = 1.0
container_height = 3.0
container_width = 4.0
nboundary_layers = 2
#h = 0.0156
h = 0.039
ro = 1000.0
co = 10.0 * np.sqrt(2*9.81*fluid_column_height)
gamma = 7.0
alpha = 0.1
beta = 0.0
B = co*co*ro/gamma
p0 = 1000.0

class DamBreak2D(Application):
    def add_user_options(self, group):
        self.hdx = 1.3
        group.add_argument(
            "--h-factor", action="store", type=float, dest="h_factor",
            default=1.0,
            help="Divide default h by this factor to change resolution"
        )

    def consume_user_options(self):
        self.h = h/self.options.h_factor
        print("Using h = %f"%self.h)
        self.hdx = self.hdx
        self.dx = self.h/self.hdx

    def configure_scheme(self):
        self.scheme.configure(h0=self.h, hdx=self.hdx)
        kernel = WendlandQuintic(dim=2)
        tf = 2.5
        from pysph.sph.integrator import TVDRK3Integrator
        self.scheme.configure_solver(
            kernel=kernel, integrator_cls=TVDRK3Integrator, tf=tf,
            adaptive_timestep=True, n_damp=50, fixed_h=False
        )

    def create_scheme(self):
        s = WCSPHScheme(
            ['fluid'], ['boundary'], dim=2, rho0=ro, c0=co,
            h0=h, hdx=1.3, gy=-9.81, alpha=alpha, beta=beta,
            gamma=gamma, hg_correction=True, update_h=True
        )
        return s

    def create_particles(self):
        geom = DamBreak2DGeometry(
            container_width=container_width, container_height=container_height,
            fluid_column_height=fluid_column_height,
            fluid_column_width=fluid_column_width, dx=self.dx, dy=self.dx,
            nboundary_layers=1, ro=ro, co=co,
            with_obstacle=False,
            beta=2.0, nfluid_offset=1, hdx=self.hdx)
        return geom.create_particles()

    def post_process(self, info_fname):
        info = self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        from pysph.solver.utils import iter_output
        t, x_max = [], []
        factor = np.sqrt(2.0*9.81/fluid_column_width)
        files = self.output_files
        for sd, array in iter_output(files, 'fluid'):
            t.append(sd['t']*factor)
            x = array.get('x')
            x_max.append(x.max())

        t, x_max = list(map(np.asarray, (t, x_max)))
        fname = os.path.join(self.output_dir, 'results.npz')
        np.savez(fname, t=t, x_max=x_max)

        import matplotlib
        matplotlib.use('Agg')
        from matplotlib import pyplot as plt
        from pysph.examples import db_exp_data as dbd
        plt.plot(t, x_max, label='Computed')
        te, xe = dbd.get_koshizuka_oka_data()
        plt.plot(te, xe, 'o', label='Koshizuka & Oka (1996)')
        plt.xlim(0, 0.7*factor); plt.ylim(0, 4.5)
        plt.xlabel('$T$'); plt.ylabel('$Z/L$')
        plt.legend(loc='upper left')
        plt.savefig(os.path.join(self.output_dir, 'x_vs_t.png'), dpi=300)
        plt.close()

if __name__ == '__main__':
    app = DamBreak2D()
    app.run()
    app.post_process(app.info_filename)



"""Tests for the simple inlet.

Copyright (c) 2015, Prabhu Ramachandran
License: BSD
"""
import numpy as np
try:
    # This is for Python-2.6.x
    import unittest2 as unittest
except ImportError:
    import unittest

from pysph.base.utils import get_particle_array
from pysph.sph.simple_inlet_outlet import SimpleInlet, SimpleOutlet


class TestSimpleInlet1D(unittest.TestCase):
    def setUp(self):
        dx = 0.1
        x = np.array([0.0])
        m = np.ones_like(x)
        h = np.ones_like(x)*dx*1.5
        p = np.ones_like(x)*5.0
        self.inlet_pa = get_particle_array(x=x, m=m, h=h, p=p)
        # Empty particle array.
        self.dest_pa = get_particle_array()
        self.dx = dx

    def test_inlet_block_has_correct_copies(self):
        # Given
        inlet = SimpleInlet(self.inlet_pa, self.dest_pa, spacing=self.dx, n=5,
                            xmin=-0.5, xmax=0.0)
        # When
        x = self.inlet_pa.x
        p = self.inlet_pa.p
        h = self.inlet_pa.h

        # Then
        self.assertEqual(len(x), 5)
        x_expect = -np.arange(0, 5)*self.dx
        self.assertListEqual(list(x), list(x_expect))
        self.assertTrue(np.allclose(p, np.ones_like(x)*5, atol=1e-14))
        self.assertTrue(np.allclose(h, np.ones_like(x)*self.dx*1.5, atol=1e-14))

        self.assertEqual(self.dest_pa.get_number_of_particles(), 0)

    def test_inlet_raises_error_on_incorrect_domain(self):
        # Given

        # When
        n = 5
        xmin = -0.4

        # Then
        self.assertRaises(RuntimeError, SimpleInlet,
                          self.inlet_pa, self.dest_pa,
                          spacing=self.dx, n=n, xmin=xmin, xmax=0.0)

    def test_update_creates_particles_in_destination(self):
        # Given
        inlet = SimpleInlet(self.inlet_pa, self.dest_pa, spacing=self.dx, n=5,
                            xmin=-0.5, xmax=0.0)
        # Two rows of particles should move out.
        self.inlet_pa.x += 0.15

        # When
        inlet.update()

        # Then
        x = self.inlet_pa.x
        p = self.inlet_pa.p
        h = self.inlet_pa.h
        self.assertEqual(len(x), 5)
        x_expect = (-np.arange(0, 5)*self.dx + 0.15)
        x_expect[x_expect > 0.0] -= 0.5
        self.assertListEqual(list(x), list(x_expect))
        self.assertTrue(np.allclose(p, np.ones_like(x)*5, atol=1e-14))
        self.assertTrue(np.allclose(h, np.ones_like(x)*self.dx*1.5, atol=1e-14))

        # The destination particle array should now have two particles.
        x = self.dest_pa.x
        p = self.dest_pa.p
        h = self.dest_pa.h
        self.assertEqual(self.dest_pa.get_number_of_particles(), 2)
        x_expect = (-np.arange(0, 2)*self.dx + 0.15)
        self.assertListEqual(list(x), list(x_expect))
        self.assertTrue(np.allclose(p, np.ones_like(x)*5, atol=1e-14))
        self.assertTrue(np.allclose(h, np.ones_like(x)*self.dx*1.5, atol=1e-14))


class TestSimpleInletGenericMotion2D(unittest.TestCase):
    def setUp(self):
        dx = 0.1
        y = np.arange(5, dtype=float)*0.1
        m = np.ones_like(y)
        h = np.ones_like(y)*dx*1.5
        p = np.ones_like(y)*5.0
        self.inlet_pa = get_particle_array(y=y, m=m, h=h, p=p)
        # Empty particle array.
        self.dest_pa = get_particle_array()
        self.dx = dx

    def test_update_correctly_handles_general_motion(self):
        # Given
        inlet_pa = self.inlet_pa
        dest_pa = self.dest_pa
        inlet = SimpleInlet(inlet_pa, dest_pa, spacing=self.dx, n=5,
                            xmin=-0.5, xmax=0.0, ymin=0.0, ymax=0.5)
        x0 = inlet_pa.x.copy()
        y0 = inlet_pa.y.copy()
        # One row and one column of particles should move out.
        dx = 0.15
        dy = 0.1
        inlet_pa.x -= dx
        inlet_pa.y -= dy

        # When
        inlet.update()

        # Then
        x = inlet_pa.x
        y = inlet_pa.y
        p = inlet_pa.p
        h = inlet_pa.h

        # The number of particles should be the same in the inlet.
        self.assertEqual(len(x), 25)
        x_expect = (x0 - dx)
        x_expect[x_expect < -0.5] += 0.5
        self.assertListEqual(list(x), list(x_expect))

        y_expect = (y0 - dy)
        y_expect[y_expect < 0.0] += 0.5
        self.assertListEqual(list(y), list(y_expect))

        self.assertTrue(np.allclose(p, np.ones_like(x)*5, atol=1e-14))
        self.assertTrue(np.allclose(h, np.ones_like(x)*self.dx*1.5, atol=1e-14))

        # The destination particle array should now have 9 particles.
        x = dest_pa.x
        y = dest_pa.y
        p = dest_pa.p
        h = dest_pa.h
        print(x, y)
        self.assertEqual(self.dest_pa.get_number_of_particles(), 9)

        # Calculate the expected positions.
        x_new = (x0 - dx)
        y_new = (y0 - dy)
        idx = np.where((x_new < -0.5) | (y_new < 0.0))[0]
        x_expect = x_new[idx]
        y_expect = y_new[idx]

        self.assertListEqual(list(x), list(x_expect))
        self.assertListEqual(list(y), list(y_expect))
        self.assertTrue(np.allclose(p, np.ones_like(x)*5, atol=1e-14))
        self.assertTrue(np.allclose(h, np.ones_like(x)*self.dx*1.5, atol=1e-14))


class TestSimpleOutlet1D(unittest.TestCase):
    def setUp(self):
        dx = 0.1
        x = np.arange(0, 10, dtype=float)*dx
        m = np.ones_like(x)
        h = np.ones_like(x)*dx*1.5
        p = np.ones_like(x)*5.0
        self.source_pa = get_particle_array(x=x, m=m, h=h, p=p)
        # Empty particle array.
        self.outlet_pa = get_particle_array()
        self.dx = dx

    def test_outlet_absorbs_particles_from_source(self):
        # Given
        outlet = SimpleOutlet(self.outlet_pa, self.source_pa, xmin=1.0,
                              xmax=1.5)

        # When
        self.source_pa.x += 0.45
        outlet.update()

        # Then
        x = self.source_pa.x
        p = self.source_pa.p
        h = self.source_pa.h
        self.assertEqual(len(x), 6)
        x_expect = np.arange(0, 6, dtype=float)*self.dx + 0.45
        self.assertListEqual(list(x), list(x_expect))
        self.assertTrue(np.allclose(p, np.ones_like(x)*5, atol=1e-14))
        self.assertTrue(np.allclose(h, np.ones_like(x)*self.dx*1.5, atol=1e-14))

        x = self.outlet_pa.x
        p = self.outlet_pa.p
        h = self.outlet_pa.h
        self.assertEqual(len(x), 4)

    def test_outlet_removes_exiting_particles_3D(self):
        # Given
        # Note that we treat the self.source_pa as the outlet in this case.
        outlet_pa = self.source_pa
        source_pa = self.outlet_pa
        outlet = SimpleOutlet(
            outlet_pa=outlet_pa, source_pa=source_pa, xmin=0,
            xmax=1.0, ymin=-0.5, ymax=0.5, zmin=-0.5, zmax=0.5
        )

        # When
        outlet_pa.x += 0.61
        outlet_pa.y[0] += 0.51
        outlet_pa.y[1] -= 0.51
        outlet_pa.z[2] += 0.51
        outlet_pa.z[3] -= 0.51
        outlet.update()

        # Then
        self.assertEqual(len(outlet_pa.x), 0)
        self.assertEqual(len(source_pa.x), 0)

if __name__ == '__main__':
    unittest.main()



import os
from os import path
try:
    from subprocess import check_output
except ImportError:
    # check_output does not exist in Python-2.6
    from subprocess import Popen, PIPE, CalledProcessError
    def check_output(*popenargs, **kwargs):
        if 'stdout' in kwargs:
            raise ValueError('stdout argument not allowed, it will be overridden.')
        process = Popen(stdout=PIPE, *popenargs, **kwargs)
        output, unused_err = process.communicate()
        retcode = process.poll()
        if retcode:
            cmd = kwargs.get("args")
            if cmd is None:
                cmd = popenargs[0]
            raise CalledProcessError(retcode, cmd, output=output)
        return output

import sys

MODE = 'normal'
if len(sys.argv) >= 2 and ('--help' in sys.argv[1:] or
        sys.argv[1] in ('--help-commands', 'egg_info', '--version',
                        'clean', 'sdist')):
    MODE = 'info'

HAVE_MPI = True
try:
    import mpi4py
except ImportError:
    HAVE_MPI = False

USE_ZOLTAN=True

compiler = 'gcc'
#compiler = 'intel'
if compiler == 'intel':
    extra_compile_args = ['-O3']
else:
    extra_compile_args = []


def get_deps(*args):
    """Given a list of basenames, this checks if a .pyx or .pxd exists
    and returns the list.
    """
    result = []
    for basename in args:
        for ext in ('.pyx', '.pxd'):
            f = basename + ext
            if path.exists(f):
                result.append(f)
    return result

def get_openmp_flags():
    """Returns any OpenMP related flags if OpenMP is avaiable on the system.
    """
    if sys.platform == 'win32':
        omp_compile_flags, omp_link_flags = ['/openmp'], []
    else:
        omp_compile_flags, omp_link_flags = ['-fopenmp'], ['-fopenmp']

    env_var = os.environ.get('USE_OPENMP', '')
    if env_var.lower() in ("0", 'false', 'n'):
        print("-"*70)
        print("OpenMP disabled by environment variable (USE_OPENMP).")
        return [], [], False

    from textwrap import dedent
    try:
        from Cython.Distutils import Extension
        from pyximport import pyxbuild
    except ImportError:
        print("Unable to import Cython, disabling OpenMP for now.")
        return [], [], False

    from distutils.errors import CompileError, LinkError
    import shutil
    import tempfile
    test_code = dedent("""
    from cython.parallel import parallel, prange, threadid
    cimport openmp
    def n_threads():
        with nogil, parallel():
            openmp.omp_get_num_threads()
    """)
    tmp_dir = tempfile.mkdtemp()
    fname = path.join(tmp_dir, 'check_omp.pyx')
    with open(fname, 'w') as fp:
        fp.write(test_code)
    extension = Extension(
        name='check_omp', sources=[fname],
        extra_compile_args=omp_compile_flags,
        extra_link_args=omp_link_flags,
    )
    has_omp = True
    try:
        mod = pyxbuild.pyx_to_dll(fname, extension, pyxbuild_dir=tmp_dir)
        print("-"*70)
        print("Using OpenMP.")
        print("-"*70)
    except CompileError:
        print("*"*70)
        print("Unable to compile OpenMP code. Not using OpenMP.")
        print("*"*70)
        has_omp = False
    except LinkError:
        print("*"*70)
        print("Unable to link OpenMP code. Not using OpenMP.")
        print("*"*70)
        has_omp = False
    finally:
        shutil.rmtree(tmp_dir)

    if has_omp:
        return omp_compile_flags, omp_link_flags, True
    else:
        return [], [], False


def get_zoltan_directory(varname):
    global USE_ZOLTAN
    d = os.environ.get(varname, '')
    if ( len(d) == 0 ):
        USE_ZOLTAN=False
        return ''
    if not path.exists(d):
        print("*"*80)
        print("%s incorrectly set to %s, not using ZOLTAN!"%(varname, d))
        print("*"*80)
        USE_ZOLTAN=False
        return ''
    return d


def get_mpi_flags():
    """Returns mpi_inc_dirs, mpi_compile_args, mpi_link_args.
    """
    global HAVE_MPI
    mpi_inc_dirs = []
    mpi_compile_args = []
    mpi_link_args = []
    if not HAVE_MPI:
        return mpi_inc_dirs, mpi_compile_args, mpi_link_args
    try:
        mpic = 'mpic++'
        if compiler == 'intel':
            link_args = check_output(
                [mpic, '-cc=icc', '-link_info'], universal_newlines=True
            ).strip()
            link_args = link_args[3:]
            compile_args = check_output(
                [mpic, '-cc=icc', '-compile_info'], universal_newlines=True
            ).strip()
            compile_args = compile_args[3:]
        else:
            link_args = check_output(
                [mpic, '--showme:link'], universal_newlines=True
            ).strip()
            compile_args = check_output(
                [mpic, '--showme:compile'], universal_newlines=True
            ).strip()
    except:
        print('-'*80)
        print("Unable to run mpic++ correctly, skipping parallel build")
        print('-'*80)
        HAVE_MPI = False
    else:
        mpi_link_args.extend(link_args.split())
        mpi_compile_args.extend(compile_args.split())
        mpi_inc_dirs.append(mpi4py.get_include())

    return mpi_inc_dirs, mpi_compile_args, mpi_link_args


def get_zoltan_args():
    """Returns zoltan_include_dirs, zoltan_library_dirs
    """
    global HAVE_MPI
    zoltan_include_dirs, zoltan_library_dirs = [], []
    if not HAVE_MPI:
        return zoltan_include_dirs, zoltan_library_dirs
    # First try with the environment variable 'ZOLTAN'
    zoltan_base = get_zoltan_directory('ZOLTAN')
    inc = lib = ''
    if len(zoltan_base) > 0:
        inc = path.join(zoltan_base, 'include')
        lib = path.join(zoltan_base, 'lib')
        if not path.exists(inc) or not path.exists(lib):
            inc = lib = ''

    # try with the older ZOLTAN include directories
    if len(inc) == 0 or len(lib) == 0:
        inc = get_zoltan_directory('ZOLTAN_INCLUDE')
        lib = get_zoltan_directory('ZOLTAN_LIBRARY')

    if (not USE_ZOLTAN):
        print("*"*80)
        print("Zoltan Environment variable not set, not using ZOLTAN!")
        print("*"*80)
        HAVE_MPI = False
    else:
        print('-'*70)
        print("Using Zoltan from:\n%s\n%s"%(inc, lib))
        print('-'*70)
        zoltan_include_dirs = [ inc ]
        zoltan_library_dirs = [ lib ]

        # PyZoltan includes
        zoltan_cython_include = [ os.path.abspath('./pyzoltan/czoltan') ]
        zoltan_include_dirs += zoltan_cython_include

    return zoltan_include_dirs, zoltan_library_dirs


def get_basic_extensions():
    if MODE == 'info':
        try:
            from Cython.Distutils import Extension
        except ImportError:
            from distutils.core import Extension
        try:
            import numpy
        except ImportError:
            include_dirs = []
        else:
            include_dirs = [numpy.get_include()]
    else:
        from Cython.Distutils import Extension
        import numpy
        include_dirs = [numpy.get_include()]

    openmp_compile_args, openmp_link_args, openmp_env = get_openmp_flags()

    ext_modules = [
        Extension(
            name="pyzoltan.core.carray",
            sources=["pyzoltan/core/carray.pyx"],
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args,
            language="c++"
        ),

        Extension(
            name="pysph.base.particle_array",
            sources=["pysph/base/particle_array.pyx"],
            depends=get_deps("pyzoltan/core/carray"),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args,
            language="c++"
        ),

        Extension(
            name="pysph.base.point",
            sources=["pysph/base/point.pyx"],
            extra_compile_args=extra_compile_args,
            include_dirs=include_dirs,
            language="c++"
        ),

        Extension(
            name="pysph.base.nnps",
            sources=["pysph/base/nnps.pyx"],
            depends=get_deps(
                "pyzoltan/core/carray", "pysph/base/point",
                "pysph/base/particle_array",
            ),
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args + openmp_compile_args,
            extra_link_args=openmp_link_args,
            cython_compile_time_env={'OPENMP': openmp_env},
            language="c++"
        ),

        # kernels used for tests
        Extension(
            name="pysph.base.c_kernels",
            sources=["pysph/base/c_kernels.pyx"],
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args,
            language="c++"
        ),

        # Eigen decomposition code
        Extension(
            name="pysph.base.linalg3",
            sources=["pysph/base/linalg3.pyx"],
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args,
            language="c++"
        ),
    ]
    return ext_modules

def get_parallel_extensions():
    if not HAVE_MPI:
        return []

    if MODE == 'info':
        from distutils.core import Extension
        include_dirs = []
        mpi_inc_dirs, mpi_compile_args, mpi_link_args = [], [], []
        zoltan_include_dirs, zoltan_library_dirs = [], []
    else:
        from Cython.Distutils import Extension
        import numpy
        include_dirs = [numpy.get_include()]
        mpi_inc_dirs, mpi_compile_args, mpi_link_args = get_mpi_flags()
        zoltan_include_dirs, zoltan_library_dirs = get_zoltan_args()

    # We should check again here as HAVE_MPI may be set to False when we try to
    # get the MPI flags and are not successful.
    if not HAVE_MPI:
        return []

    zoltan_modules = [
        Extension(
            name="pyzoltan.core.zoltan",
            sources=["pyzoltan/core/zoltan.pyx"],
            depends=get_deps(
                "pyzoltan/core/carray",
                "pyzoltan/czoltan/czoltan",
                "pyzoltan/czoltan/czoltan_types",
            ),
            include_dirs = include_dirs+zoltan_include_dirs+mpi_inc_dirs,
            library_dirs = zoltan_library_dirs,
            libraries=['zoltan', 'mpi'],
            extra_link_args=mpi_link_args,
            extra_compile_args=mpi_compile_args+extra_compile_args,
            language="c++"
        ),

        Extension(
            name="pyzoltan.core.zoltan_dd",
            sources=["pyzoltan/core/zoltan_dd.pyx"],
            depends=get_deps(
                "pyzoltan/core/carray",
                "pyzoltan/czoltan/czoltan_dd",
                "pyzoltan/czoltan/czoltan_types"
            ),
            include_dirs = include_dirs + zoltan_include_dirs + mpi_inc_dirs,
            library_dirs = zoltan_library_dirs,
            libraries=['zoltan', 'mpi'],
            extra_link_args=mpi_link_args,
            extra_compile_args=mpi_compile_args+extra_compile_args,
            language="c++"
        ),

        Extension(
            name="pyzoltan.core.zoltan_comm",
            sources=["pyzoltan/core/zoltan_comm.pyx"],
            depends=get_deps(
                "pyzoltan/core/carray",
                "pyzoltan/czoltan/zoltan_comm"
            ),
            include_dirs = include_dirs + zoltan_include_dirs + mpi_inc_dirs,
            library_dirs = zoltan_library_dirs,
            libraries=['zoltan', 'mpi'],
            extra_link_args=mpi_link_args,
            extra_compile_args=mpi_compile_args+extra_compile_args,
            language="c++"
        ),
    ]

    parallel_modules = [

        Extension(
            name="pysph.parallel.parallel_manager",
            sources=["pysph/parallel/parallel_manager.pyx"],
            depends=get_deps("pyzoltan/core/carray",
                "pyzoltan/core/zoltan", "pyzoltan/core/zoltan_comm",
                "pysph/base/point", "pysph/base/particle_array",
                "pysph/base/nnps"
            ),
            include_dirs = include_dirs + mpi_inc_dirs + zoltan_include_dirs,
            library_dirs = zoltan_library_dirs,
            libraries = ['zoltan', 'mpi'],
            extra_link_args=mpi_link_args,
            extra_compile_args=mpi_compile_args+extra_compile_args,
            language="c++"
        ),
    ]
    return zoltan_modules + parallel_modules


def create_sources():
    if 'build_ext' in sys.argv or 'develop' in sys.argv or 'install' in sys.argv:
        generator = path.join('pyzoltan', 'core', 'generator.py')
        for pth in (path.join('pyzoltan', 'core'), path.join('pysph', 'base')):
            cmd = [sys.executable, generator, path.abspath(pth)]
            print(check_output(cmd).decode())


def setup_package():
    from setuptools import find_packages, setup
    if MODE == 'info':
        cmdclass = {}
    else:
        from Cython.Distutils import build_ext
        cmdclass = {'build_ext': build_ext}

    create_sources()

    # Extract the version information from pysph/__init__.py
    info = {}
    module = path.join('pysph', '__init__.py')
    exec(compile(open(module).read(), module, 'exec'), info)

    # The requirements.
    install_requires = ['numpy', 'mako', 'Cython>=0.20', 'setuptools>=3.2']
    if sys.version_info[:2] == (2, 6):
        install_requires += [
            'ordereddict', 'importlib', 'unittest2'
        ]

    extras_require = dict(
        mpi=['mpi4py>=1.2'],
        ui=['mayavi>=4.0'],
        test=['nose>=1.0.0', 'mock>=1.0']
    )

    everything = set()
    for dep in extras_require.values():
        everything.update(dep)
    extras_require['all'] = everything

    ext_modules = get_basic_extensions() + get_parallel_extensions()


    setup(name='PySPH',
          version = info['__version__'],
          author = 'PySPH Developers',
          author_email = 'pysph-dev@googlegroups.com',
          description = "A general purpose Smoothed Particle Hydrodynamics framework",
          long_description = open('README.rst').read(),
          url = 'http://pysph.bitbucket.org',
          license = "BSD",
          keywords = "SPH simulation computational fluid dynamics",
          test_suite = "nose.collector",
          packages = find_packages(),
          package_data = {
              '': ['*.pxd', '*.mako', '*.txt.gz', '*.txt', '*.vtk.gz', '*.gz',
                  '*.rst', 'ndspmhd-sedov-initial-conditions.npz']
          },
          # exclude package data in installation.
          exclude_package_data = {
              '' : ['Makefile', '*.bat', '*.cfg', '*.rst', '*.sh', '*.yml'],
          },
          ext_modules = ext_modules,
          include_package_data = True,
          cmdclass = cmdclass,
          install_requires = install_requires,
          extras_require = extras_require,
          zip_safe = False,
          entry_points = """
              [console_scripts]
              pysph = pysph.tools.cli:main
              """,
          platforms=['Linux', 'Mac OS-X', 'Unix', 'Windows'],
          classifiers = [c.strip() for c in """\
            Development Status :: 3 - Alpha
            Environment :: Console
            Intended Audience :: Developers
            Intended Audience :: Science/Research
            License :: OSI Approved :: BSD License
            Natural Language :: English
            Operating System :: MacOS :: MacOS X
            Operating System :: Microsoft :: Windows
            Operating System :: POSIX
            Operating System :: Unix
            Programming Language :: Python
            Programming Language :: Python :: 3
            Topic :: Scientific/Engineering
            Topic :: Scientific/Engineering :: Physics
            Topic :: Software Development :: Libraries
            """.splitlines() if len(c.split()) > 0],
          )

if __name__ == '__main__':
    setup_package()



""" This simulates two square blocks of water colliding with each other.
"""

import sys
import os
# Need this to import db_geometry.
sys.path.append(os.pardir)
import numpy
from db_geometry import create_2D_filled_region

from pysph.base.kernels import CubicSpline
from pysph.base.utils import get_particle_array_iisph
from pysph.sph.equation import Group

from pysph.sph.iisph import (AdvectionAcceleration, ComputeAII, ComputeDII,
    ComputeDIJPJ, ComputeRhoAdvection, IISPHStep, PressureSolve,
    PressureForce, SummationDensity, ViscosityAcceleration)

from pysph.sph.integrator import EulerIntegrator


from pysph.solver.application import Application
from pysph.solver.solver import Solver

dx = 0.025
hdx = 1.0
rho0 = 1000

def create_particles():
    x1, y1 = create_2D_filled_region(-1, 0, 0, 1, dx)
    x2, y2 = create_2D_filled_region(0.5, 0, 1.5, 1, dx)

    x = numpy.concatenate((x1, x2))
    y = numpy.concatenate((y1, y2))
    u1 = numpy.ones_like(x1)
    u2 = -numpy.ones_like(x2)
    u = numpy.concatenate((u1, u2))

    rho = numpy.ones_like(u)*rho0
    h = numpy.ones_like(u)*hdx*dx
    m = numpy.ones_like(u)*dx*dx*rho0

    pa = get_particle_array_iisph(
        name='fluid', x=x, y=y, u=u, rho=rho, m=m, h=h
    )
    return [pa,]


dim = 2
dt = 1e-2
tf = 1.0

# Create the application.
app = Application()

# Create the kernel
#kernel = Gaussian(dim=dim)
kernel = CubicSpline(dim=dim)


# Create the Integrator. Currently, PySPH supports multi-stage,
# predictor corrector and a TVD-RK3 integrators.

integrator = EulerIntegrator(fluid=IISPHStep())

# Create a solver.
solver = Solver(kernel=kernel, dim=dim, integrator=integrator,
                dt=dt, tf=tf, adaptive_timestep=True,
                fixed_h=False)
solver.set_print_freq(10)

# create the equations
equations = [

    Group(
        equations=[
            SummationDensity(dest='fluid', sources=['fluid']),
        ],
        real=False
    ),

    Group(
        equations=[
            AdvectionAcceleration(
                dest='fluid', sources=None, gx=0.0, gy=0.0, gz=0.0
            ),
            #ViscosityAcceleration(dest='fluid', sources=['fluid'], nu=8.9e-4),
            ComputeDII(dest='fluid', sources=['fluid']),
        ]
    ),

    Group(
        equations=[
            ComputeRhoAdvection(dest='fluid', sources=['fluid']),
            ComputeAII(dest='fluid', sources=['fluid']),
        ]
    ),

    #####################################################################
    # "Pressure solve" step as per algorithm 1 in paper.
    Group(
        equations=[
            Group(
                equations=[
                    ComputeDIJPJ(dest='fluid', sources=['fluid']),
                ]
            ),
            Group(
                equations=[
                    PressureSolve(
                        dest='fluid', sources=['fluid'], rho0=rho0,
                        tolerance=1e-3, debug=False
                    ),
                  ]
            ),
        ],
        iterate=True,
        max_iterations=20,
        min_iterations=2
    ),

    Group(
        equations=[
            PressureForce(dest='fluid', sources=['fluid']),
        ],
    ),
]

# Setup the application and solver.  This also generates the particles.
app.setup(solver=solver, equations=equations,
          particle_factory=create_particles)

app.run()



"""Tests for the PySPH parallel module"""


import os
import shutil
import tempfile
import unittest
from nose.plugins.attrib import attr
from pysph.tools import run_parallel_script

run_parallel_script.skip_if_no_mpi4py()
path = run_parallel_script.get_directory(__file__)


class ParticleArrayExchangeTestCase(unittest.TestCase):

    @attr(parallel=True)
    def test_lb_exchange(self):
        run_parallel_script.run(filename='lb_exchange.py', nprocs=4, path=path)

    @attr(parallel=True)
    def test_remote_exchange(self):
        run_parallel_script.run(filename='remote_exchange.py', nprocs=4, path=path)


class SummationDensityTestCase(unittest.TestCase):

    @attr(slow=True, parallel=True)
    def test_summation_density(self):
        run_parallel_script.run(filename='summation_density.py', nprocs=4,
                                path=path)


class MPIReduceArrayTestCase(unittest.TestCase):

    def setUp(self):
        self.root = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.root)

    @attr(parallel=True)
    def test_mpi_reduce_array(self):
        run_parallel_script.run(
            filename='reduce_array.py', nprocs=4, path=path
        )

    @attr(parallel=True)
    def test_parallel_reduce(self):
        args = ['--directory=%s'%self.root]
        run_parallel_script.run(
            filename='simple_reduction.py', args=args, nprocs=4, path=path
        )


class DumpLoadTestCase(unittest.TestCase):

    @attr(parallel=True)
    def test_dump_and_load_work_in_parallel(self):
        run_parallel_script.run(
            filename='check_dump_load.py', nprocs=4, path=path
        )


if __name__ == '__main__':
    unittest.main()



"""Wrapper for the Zoltan Unstructured Communication Package

The Unstructured Communication utilities simplifies some common
point-to-point message passing paradigms required for dynamic
applications. 

The main wrapper object for this package is the ZComm object which is
described below.

"""
cimport mpi4py.MPI as mpi
from mpi4py cimport mpi_c as mpic

# Cython for pure mode
cimport cython

# NUMPY
import numpy as np
cimport numpy as np

# Zoltan error checking
from zoltan cimport _check_error

cdef class ZComm:
    """Wrapper for simplified unstructured point-to-point communication

    The ZComm wrapper is used when each processor knows a list of
    objects it must send to other processors but does not know of what
    objects it must receive from them. It is designed to work with
    NumPy arrays for convenience.

    Processor 0 sends data (uint32) [4, 9, 6] to processors [1, 2, 2]
    Processor 1 sends data (uint32) [5, 3] to processors [0, 0]
    Processor 2 sends data (uint32) [1, 9, 0, 2] to processors [0, 1, 1, 0]

    Note that the size of data to be sent is different on each
    processor. Each processor instantiates the data and the ZComm
    object like so:

    >>> sendbuf = numpy.array( [...], dtype=numpy.uint32 )
    >>> proclist = numpy.array( [...], dtype=numpy.int32 )
    >>> nsend = sendbuf.size

    >>> zcomm = ZComm(comm, tag=0, nsend=nsend, proclist=proclist)
    
    """
    def __init__(self, object comm, int tag, int nsend, np.ndarray proclist):
        """Constructor for the ZComm object

        Parameters:
        
        comm : MPI Comm
            MPI communicator (mpi4py.MPI.COMM_WORLD)

        tag : int
            Message tag for the unstructured communication

        nsend : int
            Number of objects this processor has to send

        proclist : numpy.ndarray
            Array of size 'nsend' indicating where each object in the 
            sendbuf (to be defined) will be sent.

        Notes:
        
        In the general case, every processor has some data it knows it
        must share with remote processors. 

        As an example, we may have that processor 2 sends 4 ojects of
        data to processors to [0, 3, 1, 0]. In this example, the first
        object is sent to processor 0, the second to processor 3 and
        so on and so forth. The ZComm for this processor can be
        constructed as

        >>> nsend = 4
        >>> proclist = numpy.array( [0, 3, 1, 0], dtype=nupmy.int32 )
        >>> zcomm = ZComm(mpi.COMM_WORLD, nsend=nsend, tag=0, proclist=proclist)

        Upon instantiation, the zcomm object will know of the number
        of objects it must receive from all remote processors:

        >>> nrecv = zcomm.nreturn

        Knowing the number of objects that must be received, we can
        allocate buffers of the appropriate size to collect this
        data. The actual transfer of data is effected with a call to
        the `ZComm.Comm_Do` method

        We can use the same ZComm object for multiple exchanges of
        data of different types as long as the number of objects to be
        exchanged are the same. For example, a processor may want to
        exchange solution data (doubles) along with indices
        (uints). The same plan can be used in this case with the
        provision to alter the number of bytes per object with the
        method `ZComm.set_nbytes`
        
        """
        self.comm = comm
        self.rank = comm.Get_rank()
        self.size = comm.Get_size()

        self.tag = tag
        self.nsend = nsend
        self.proclist = proclist.astype(np.int32)

        self.nreturn = 0

        # the size of each element to exchange
        self.nbytes = 8

        # internally call Zoltan_Comm_Create
        self.initialize()

    def __dealloc__(self):
        "Cleanup"
        cdef zcomm.ZOLTAN_COMM_OBJ* _zoltan_comm_obj = self._zoltan_comm_obj
        zcomm.Zoltan_Comm_Destroy(&_zoltan_comm_obj)

    def initialize(self):
        """Zoltan_Comm_Create

        This function calls the Zoltan_Comm_Create function to create
        the unstructured communication plan. This plan can
        subsequently be used to effect data transfers in an
        unstructured manner between processors.

        Upon return, each processor knows the number of objects that
        are to be received through the data attribute `nreturn`
        
        """
        cdef zcomm.ZOLTAN_COMM_OBJ** zoltan_comm_obj = &self._zoltan_comm_obj

        cdef np.ndarray[ndim=1, dtype=np.int32_t] _proclist = self.proclist
        cdef int* proclist = <int*>_proclist.data

        cdef mpi.Comm _comm = self.comm
        cdef mpic.MPI_Comm comm = _comm.ob_mpi

        cdef int nsend = self.nsend, tag = self.tag

        # cython declarations
        cdef int ierr, _nreturn = -1
        cdef int* nreturn = &_nreturn

        # create the communication plan
        ierr = zcomm.Zoltan_Comm_Create(zoltan_comm_obj,
                                        nsend,
                                        proclist,
                                        comm,
                                        tag,
                                        nreturn)

        # check for the error
        _check_error(ierr)

        # save the number of objects to be returned
        self.nreturn = _nreturn

    def Comm_Do(self, np.ndarray _sendbuf, np.ndarray _recvbuf):
        """Perform an unstructured communication between processors

        Parameters:
        
        _sendbuf : np.ndarray
            The array of data to be sent by this processor

        _recvbuf : np.ndarray
            The array of data to be received by this processor

        Notes:
        
        Internally, Zoltan_Comm_Do accepts char* buffers to move the
        data between processors. The number of objects is determined
        by the `nbytes` argument. 

        The `nsend` argument used to create the ZComm object and the
        `nbytes` argument should be consistent to avoid strange
        behaviour.

        """
        cdef zcomm.ZOLTAN_COMM_OBJ* _zoltan_comm_obj = self._zoltan_comm_obj
        cdef char* send_data = _sendbuf.data
        cdef char* recv_data = _recvbuf.data
        cdef int ierr, tag = self.tag, nbytes = self.nbytes

        ierr = zcomm.Zoltan_Comm_Do(
            _zoltan_comm_obj,
            tag,
            send_data,
            nbytes,
            recv_data)

        _check_error(ierr)

    def Comm_Do_Reverse(self, np.ndarray _sendbuf, np.ndarray recvbuf):
        """Perform the reverse of the unstructured communication
        between processors

        Parameters:
        
        _sendbuf : np.ndarray
            The array of data to be sent by this processor

        Notes:
        
        Internally, Zoltan_Comm_Do accepts char* buffers to move the
        data between processors. The number of objects is determined
        by the `nbytes` argument. 

        """
        cdef zcomm.ZOLTAN_COMM_OBJ* _zoltan_comm_obj = self._zoltan_comm_obj
        cdef char* send_data = _sendbuf.data
        cdef int ierr, tag = self.tag, nbytes = self.nbytes

        # the returned data will be updated. The number of objects to
        # send for Do_Reverse is therefore equal to nreturn
        cdef int nsend = self.nreturn

        # sizes pointer is null for equal sized objects
        cdef int* sizesp = NULL

        #cdef np.ndarray recvbuf = np.zeros( self.nsend, dtype=dtype )
        cdef char* _recvbuf = recvbuf.data

        # Zoltan interface function
        ierr = zcomm.Zoltan_Comm_Do_Reverse(
            _zoltan_comm_obj,
            tag,
            send_data,
            nbytes,
            sizesp,
            _recvbuf)

        _check_error(ierr)

        #return recvbuf

    def set_nbytes(self, int nbytes, object dtype=None):
        "Set the number of bytes for each object"
        self.nbytes = nbytes
        self.dtype = dtype

    def set_tag(self, int tag):
        "Set the message tag for this plan"
        self.tag = tag



from contextlib import contextmanager
from distutils.sysconfig import get_config_var
import os
from os.path import join, exists
import shutil
import sys
import tempfile
from textwrap import dedent
from multiprocessing import Pool
from unittest import TestCase, main

try:
    from unittest import mock
except ImportError:
    import mock

from pysph.base.ext_module import get_md5, ExtModule


def _check_write_source(root):
    """Used to create an ExtModule and test if a file was opened.

    It returns the number of times "open" was called.
    """
    m = mock.mock_open()
    with mock.patch('pysph.base.ext_module.open', m, create=True):
        s = ExtModule("print 'hello'", root=root)
    if m.called:
        with open(*m.call_args[0]) as fp:
            fp.write("junk")
    return m.call_count

def _check_compile(root):
    with mock.patch('shutil.copy') as m:
        s = ExtModule("print 'hello'", root=root)
        s.build()
    if m.called:
        # If it was called, do the copy to mimic the action.
        shutil.copy(*m.call_args[0])
    return m.call_count


class TestMiscExtMod(TestCase):
    def test_md5(self):
        data = "hello world"
        # Two calls with same data produce same result
        self.assertEqual(get_md5(data), get_md5(data))
        # Two calls with different data produce different md5sums.
        self.assertNotEqual(get_md5(data), get_md5(data + ' '))


class TestExtModule(TestCase):
    def setUp(self):
        self.root = tempfile.mkdtemp()
        self.data = dedent('''\
        def f():
            return "hello world"
        ''')

    def tearDown(self):
        if sys.platform.startswith('win'):
            from exceptions import WindowsError
            try:
                shutil.rmtree(self.root)
            except WindowsError:
                pass
        else:
            shutil.rmtree(self.root)

    def test_constructor(self):
        data = self.data
        s = ExtModule(data, root=self.root)
        self.assertTrue(exists(join(self.root, 'build')))

        self.assertEqual(s.hash, get_md5(data))
        self.assertEqual(s.code, data)
        expect_name = 'm_%s'%(s.hash)
        self.assertEqual(s.name, expect_name)
        self.assertEqual(s.src_path, join(self.root, expect_name +'.pyx'))
        self.assertEqual(s.ext_path,
                         join(self.root, expect_name + get_config_var('SO')))

        self.assertTrue(exists(s.src_path))
        self.assertEqual(data, open(s.src_path).read())

    def test_default_root(self):
        try:
            data = self.data
            s = ExtModule(data)
            self.assertTrue(exists(join(s.root, 'build')))
            self.assertEqual(s.hash, get_md5(data))
            self.assertEqual(s.code, data)
            self.assertTrue(exists(s.src_path))
            self.assertEqual(data, open(s.src_path).read())
        finally:
            os.unlink(s.src_path)

    def test_load_module(self):
        data = self.data
        s = ExtModule(data, root=self.root)
        mod = s.load()
        self.assertEqual(mod.f(), "hello world")
        self.assertTrue(exists(s.ext_path))

    def _create_dummy_module(self):
        code = "def hello(): return 'hello'"
        modname = 'test_rebuild.py'
        f = join(self.root, modname)
        with open(f, 'w') as fp:
            fp.write(code)
        return f

    @contextmanager
    def _add_root_to_sys_path(self):
        import sys
        if self.root not in sys.path:
            sys.path.insert(0, self.root)
        try:
            yield
        finally:
            sys.path.remove(self.root)

    def test_rebuild_when_dependencies_change(self):
        # Given.
        data = self.data
        depends = ["test_rebuild"]
        s = ExtModule(data, root=self.root, depends=depends)
        fname = self._create_dummy_module()
        f_stat = os.stat(fname)

        with self._add_root_to_sys_path():
            # When
            self.assertTrue(s.should_recompile())
            s.build()

            # Then.
            self.assertFalse(s.should_recompile())

            # Now lets re-create the module and try again.

            # When.
            fname = self._create_dummy_module()
            # Update the timestamp to make it newer, otherwise we need to
            # sleep.
            os.utime(fname, (f_stat.st_atime, f_stat.st_mtime + 10))

            # Then.
            self.assertTrue(s.should_recompile())

    def test_that_multiple_writes_do_not_occur_for_same_source(self):
        # Given
        n_proc = 5
        p = Pool(n_proc)

        # When

        # Note that _create_extension cannot be defined here or even in the
        # class as a nested function or instance method cannot be pickled.

        result = p.map(_check_write_source, [self.root]*n_proc)
        p.close()

        # Then
        # The file should have been opened only once.
        self.assertEqual(sum(result), 1)

    def test_that_multiple_compiles_do_not_occur_for_same_source(self):
        # Given
        n_proc = 5
        p = Pool(n_proc)

        # When

        # Note that _check_compile cannot be defined here or even in the
        # class as a nested function or instance method cannot be pickled.

        result = p.map(_check_compile, [self.root]*n_proc)
        p.close()

        # Then
        # The shutil.copy should have been run only once.
        self.assertEqual(sum(result), 1)


if __name__ == '__main__':
    main()



"""A convenience class that combines an AccelerationEval and an SPHCompiler to
allow a user to specify particle arrays, equations, an optional domain and
kernel to produce an SPH evaluation.

This is handy for post-processing.

"""

from pysph.base.kernels import Gaussian
from pysph.base.nnps import LinkedListNNPS as NNPS
from pysph.sph.acceleration_eval import AccelerationEval
from pysph.sph.sph_compiler import SPHCompiler


class SPHEvaluator(object):
    def __init__(self, arrays, equations, dim, kernel=None,
                 domain_manager=None):
        """Constructor.

        Parameters
        ----------
        arrays: list(ParticleArray)
        equations: list
        dim: int
        kernel: kernel instance.
        domain_manager: DomainManager
        """
        self.arrays = arrays
        self.equations = equations
        self.domain_manager = domain_manager
        self.dim = dim
        if kernel is None:
            self.kernel = Gaussian(dim=dim)
        else:
            self.kernel = kernel

        self.func_eval = AccelerationEval(arrays, equations, self.kernel)
        compiler = SPHCompiler(self.func_eval, None)
        compiler.compile()
        self._create_nnps(arrays)

    def evaluate(self, t=0.0, dt=0.1):
        """Evalute the SPH equations, dummy t and dt values can
        be passed.
        """
        self.func_eval.compute(t, dt)

    def update_particle_arrays(self, arrays):
        self._create_nnps(arrays)
        self.func_eval.update_particle_arrays(arrays)

    #### Private protocol ###################################################
    def _create_nnps(self, arrays):
        self.nnps = NNPS(dim=self.kernel.dim, particles=arrays,
                         radius_scale=self.kernel.radius_scale,
                         domain=self.domain_manager,
                         cache=True)
        self.nnps.update()
        self.func_eval.set_nnps(self.nnps)



from collections import defaultdict
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

from pysph.sph.equation import Group, get_arrays_used_in_equation


###############################################################################
def group_equations(equations):
    """Checks the given equations and ensures the following:

     - Raises an error if the user  mixes Groups and Equations.

     - If only equations are given as a list, return a single group with all
       these equations.
    """
    only_groups = [x for x in equations if isinstance(x, Group)]
    if len(only_groups) > 0 and len(only_groups) != len(equations):
        raise ValueError('All elements must be Groups if you use groups.')
    if len(only_groups) == 0:
        return [Group(equations)]
    else:
        return equations


###############################################################################
def check_equation_array_properties(equation, particle_arrays):
    """Given an equation and the particle arrays, check if the particle arrays
    have the necessary properties.
    """
    #p_arrays = {x.name:x for x in particle_arrays}
    p_arrays = dict((x.name, x) for x in particle_arrays)
    _src, _dest = get_arrays_used_in_equation(equation)
    eq_src = set([x[2:] for x in _src])
    eq_dest = set([x[2:] for x in _dest])

    def _check_array(array, eq_props, errors):
        """Updates the `errors` with any errors.
        """
        props = set(list(array.properties.keys()) +
                    list(array.constants.keys()))
        if not eq_props < props:
            errors[array.name].update(eq_props - props)

    errors = defaultdict(set)
    _check_array(p_arrays[equation.dest], eq_dest, errors)
    if equation.sources is not None:
        for src in equation.sources:
            _check_array(p_arrays[src], eq_src, errors)

    if len(errors) > 0:
        msg = "ERROR: Missing array properties for equation: %s\n"%equation.name
        for name, missing in errors.items():
            msg += "Array '%s' missing properties %s.\n"%(name, missing)
        print(msg)
        raise RuntimeError(msg)


###############################################################################
class MegaGroup(object):
    """A mega-group refactors actual equation Groups into a more
    organized form as described below.  They inherit all properties of the
    Group so these can be used while generating code but delegate the tasks to
    real groups underneath that are assembled from the original group.

    MegaGroups are organized as:
        {destination: (eqs_with_no_source, sources, all_eqs)}
        eqs_with_no_source: Group([equations]) all SPH Equations with no source.
        sources are {source: Group([equations...])}
        all_eqs is a Group of all equations having this destination.

    This is what is stored in the `data` attribute.
    """
    def __init__(self, group):
        self._orig_group = group
        self._copy_props(group)
        self.data = self._make_data(group)

    def get_converged_condition(self):
        return self._orig_group.get_converged_condition()

    def _copy_props(self, group):
        for key in ('real', 'update_nnps', 'iterate',
                    'max_iterations', 'min_iterations', 'has_subgroups'):
            setattr(self, key, getattr(group, key))

    def _make_data(self, group):
        equations = group.equations

        if group.has_subgroups:
            return [MegaGroup(g) for g in equations]

        dest_list = []
        for equation in equations:
            dest = equation.dest
            if dest not in dest_list:
                dest_list.append(dest)

        dests = OrderedDict()
        for dest in dest_list:
            sources = defaultdict(list)
            eqs_with_no_source = [] # For equations that have no source.
            all_eqs = set()
            for equation in equations:
                if equation.dest != dest:
                    continue
                all_eqs.add(equation)
                if equation.no_source:
                    eqs_with_no_source.append(equation)
                else:
                    for src in equation.sources:
                        sources[src].append(equation)

            for src in sources:
                eqs = sources[src]
                sources[src] = Group(eqs)

            # Sort the all_eqs set; so the order is deterministic.  Without
            # this a  user may get a recompilation for no obvious reason.
            all_equations = list(all_eqs)
            all_equations.sort(key=lambda x:x.__class__.__name__)
            dests[dest] = (Group(eqs_with_no_source), sources,
                           Group(all_equations))

        return dests


###############################################################################
class AccelerationEval(object):
    def __init__(self, particle_arrays, equations, kernel, mode='serial'):
        """

        Parameters
        ----------

        particle_arrays: list(ParticleArray): list of particle arrays to use.
        equations: list: A list of equations/groups.
        kernel: The kernel to use.
        parallel: str: One of 'serial', 'mpi'.
        """
        self.particle_arrays = particle_arrays
        self.equation_groups = group_equations(equations)
        self.kernel = kernel
        self.nnps = None
        self.mode = mode

        all_equations = []
        for group in self.equation_groups:
            if group.has_subgroups:
                for g in group.equations:
                    all_equations.extend(g.equations)
            else:
                all_equations.extend(group.equations)
        self.all_group = Group(equations=all_equations)

        for equation in all_equations:
            check_equation_array_properties(equation, particle_arrays)

        self.mega_groups = [MegaGroup(g) for g in self.equation_groups]
        self.c_acceleration_eval = None

    ##########################################################################
    # Public interface.
    ##########################################################################
    def compute(self, t, dt):
        """Compute the accelerations given the current time, t, and the
        timestep, dt.
        """
        self.c_acceleration_eval.compute(t, dt)

    def set_compiled_object(self, c_acceleration_eval):
        """Set the high-performance compiled object to call internally.
        """
        self.c_acceleration_eval = c_acceleration_eval

    def set_nnps(self, nnps):
        self.nnps = nnps
        self.c_acceleration_eval.set_nnps(nnps)

    def update_particle_arrays(self, particle_arrays):
        """Call this to update the particle arrays with new ones.  Make sure
        though that the same properties exist in both or you will get a
        segfault.
        """
        self.c_acceleration_eval.update_particle_arrays(particle_arrays)
